#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h> /* ntohl() htonl() */

#include <common.h>

#include <hash.h>
#include <dbgmem.h>
#include <file.h>

#define FILENAMELEN 40

struct fileinfo {
  char filename[FILENAMELEN];
  long pos;
  long size;
};

/* File format: 
   all longs are stored in network format (big endian)
   all chars are 8 bit ascii

file:
 long numfiles
 
 1..numfiles *
  char name[FILENAMELEN]
  long pos
  long size
  long options
*/

#define FILE_SIZEOFMEM (FILENAMELEN+2*4)

char *FileFI2Mem(struct fileinfo fileinfo)
{
 static char buf[FILE_SIZEOFMEM];
 long data;
 
 memcpy(buf,fileinfo.filename,FILENAMELEN);
 data=htonl(fileinfo.pos);
 memcpy(buf+FILENAMELEN,&data,4);
 data=htonl(fileinfo.size);
 memcpy(buf+FILENAMELEN+4,&data,4);

 return buf;
}

struct fileinfo FileMem2FI(char *buf)
{
 struct fileinfo fi;
 long data;
 
 memcpy(fi.filename,buf,FILENAMELEN);
 
 memcpy(&data,buf+FILENAMELEN,4);
 fi.pos=ntohl(data);
 memcpy(&data,buf+FILENAMELEN+4,4);
 fi.size=ntohl(data);

 return fi;
}

static long FileNumFiles=0;
static struct fileinfo *FileInfos=0;
static int FileDataFile=0;
static hashtable FileHashTable=0;

void FileClose(void)
{
  if (FileDataFile!=0) {
    close(FileDataFile);
    FileDataFile=0;
  }
  if (FileInfos!=0) {
    free(FileInfos);
    FileInfos=0;
  }
  if (FileHashTable!=0) {
    HashDestroy(FileHashTable);
    FileHashTable=0;
  }
  FileNumFiles=0;

}

int FileOpen(char *datafilename)
{
  long data,i;
  char buf[FILE_SIZEOFMEM];
  FileDataFile=open(datafilename,OPEN_READ_BINARY);
  if (FileDataFile<0) {
    return 1;
  }
  
  read(FileDataFile,&data,4);
  FileNumFiles=ntohl(data);

  FileInfos=malloc(sizeof(struct fileinfo)*FileNumFiles);
  if (FileInfos==0) {
    return 1;
  }
  
  FileHashTable=HashCreate(FileNumFiles*2);
  
  for (i=0;i<FileNumFiles;i++) {
    read(FileDataFile,buf,FILE_SIZEOFMEM);
    FileInfos[i]=FileMem2FI(buf);
    HashInsert(FileHashTable,&FileInfos[i],FileInfos[i].filename);
  }

  return 0;
}

void *FileLoad(char *filename)
{
  struct fileinfo *file;
  void *ptr;
  long readsize;
 

  file=HashLookup(FileHashTable,filename);

  if (file==0) {
    return 0;
  }

  ptr=malloc(file->size);
  if (ptr==0) {
    return 0;
  }

  lseek(FileDataFile,file->pos,SEEK_SET);

  readsize=read(FileDataFile,ptr,file->size);
  if (readsize!=file->size) {
    free(ptr);
    return 0;
  }

  return ptr;
}

long FileSize(char *filename)
{
  struct fileinfo *file;

  file=HashLookup(FileHashTable,filename);

  if (file==0) {
    return 0;
  }

  return file->size;
}


static int FileCurrent;

char *FileListFirst(void)
{
  FileCurrent=0;

  if (FileNumFiles<1) {
    return 0;
  } else {
    return FileInfos[FileCurrent].filename;
  }
}

char *FileListNext(void)
{
  FileCurrent++;

  if (FileCurrent>=FileNumFiles) {
    return 0;
  } else {
    return FileInfos[FileCurrent].filename;
  }
}

/*********** For Creating Data Files: **************/

struct fileinfolist {
  struct fileinfo *fileinfo;
  struct fileinfolist *next;
};

struct fileinfolist *FileInfoList=0;
int FileNumNewFiles=0;

int FileNew(void)
{

  FileNumNewFiles=0;

  return 0;
}

int FileAdd(char *filename)
{
  struct fileinfolist *new;
  struct stat s;
  long size;

  new=malloc(sizeof(struct fileinfolist));
  if (new==0) {
    return 1;
  }
  new->fileinfo=malloc(sizeof(struct fileinfo));
  if (new->fileinfo==0) {
    free(new);
    return 1;
  }


  if (stat(filename,&s)) {
    free(new->fileinfo);
    free(new);
    return 1;
  }
  size=s.st_size;
  if (size==0) {
    free(new->fileinfo);
    free(new);
    return 1;
  }

  new->fileinfo->size=size;
  new->fileinfo->pos=0;
  strncpy(new->fileinfo->filename,filename,FILENAMELEN);


  FileNumNewFiles++;

  new->next=FileInfoList;
  FileInfoList=new;

  return 0;
}

int FileAppend(int to,int from)
{
  char *ptr;
  int size;

  ptr=malloc(1024);
  if (ptr==0) {
    return 1;
  }

  do {
    size=read(from,ptr,1024);
    write(to,ptr,size);
  } while (size>0);

  free(ptr);
		
  return 0;
}



int FileCreate(char *datafilename)
{
  struct fileinfolist *info,*old;
  int datafile;
  int fd;
  long data;
  long offset;

  offset=4+FILE_SIZEOFMEM*FileNumNewFiles;
 
  info=FileInfoList;
  while(info!=0) {
    info->fileinfo->pos=offset;
    offset=offset+info->fileinfo->size;
    info=info->next;
  }

  datafile=
    open(datafilename,OPEN_WRITE_BINARY,S_IREAD|S_IWRITE);

  if (datafile==0) {
    return 1;
  }

  data=htonl(FileNumNewFiles);
  write(datafile,&data,4);

  info=FileInfoList;
  while(info!=0) {
    write(datafile,FileFI2Mem(*info->fileinfo),FILE_SIZEOFMEM);
    info=info->next;
  }

  info=FileInfoList;
  while(info!=0) {
    fd=open(info->fileinfo->filename,OPEN_READ_BINARY);
    if (fd<0) {
      return 1;
    }
    FileAppend(datafile,fd);
    close(fd);
    info=info->next;
  }
 
  close(datafile);

  info=FileInfoList;
  while(info!=0) {
    free(info->fileinfo);
    old=info;
    info=info->next;
    free(old);
  }
  
  return 0;
}
